<?php

/**
 * Class SYSP_Image_Check
 * @version  1.0.0
 *
 */

defined( 'ABSPATH' ) || exit;

class SYSP_Image_Check {

	private $version, $settings;

	/**
	 * SYSP_Image_Check constructor.
	 */
	public function __construct( $version ) {

		$this->version = $version;

		$this->settings = get_option( 'syncspider_settings_woo' );

//		// after product is created / updated
		add_action( 'woocommerce_rest_insert_product_object', array( $this, 'sysp_update_product_images' ), 10, 3 );
		add_action( 'woocommerce_rest_insert_product_variation_object', array(
			$this,
			'sysp_update_variation_image'
		), 10, 3 );
	}


	/**
	 * Update product images if _sysp_bulk_images meta is set. Check for duplicates only on current product (attached + selected for gallery
	 *
	 * @param WC_Data $object Inserted object.
	 * @param WP_REST_Request $request Request object.
	 * @param boolean $creating True when creating object, false when updating.
	 *
	 * @return mixed
	 */
	public function sysp_update_product_images( $object, $request, $creating ) {
		global $wpdb;
		$new_images       = $object->get_meta( '_sysp_bulk_images' );
		$final_images_ids = [];

		if ( $new_images ) {
			$all_imgs_ids = [];
			$all_ids      = [];
			$gallery_ids  = [];

			// all variation ids
			if ( $object->get_type() == 'variable' ) {
				$all_ids = $object->get_children();
			}
			$all_ids[] = $object->get_id();

			// get ids of gallery images + featured image. Doing it with db query to save memory
			$image_rows = $wpdb->get_results( "SELECT `meta_value` FROM `{$wpdb->prefix}postmeta` WHERE (`meta_key` = '_thumbnail_id' OR `meta_key` = '_product_image_gallery') AND `post_id` in (" . implode( ',', $all_ids ) . ");", ARRAY_A );

			if ( $image_rows ) {
				$gallery_ids  = implode( ',', array_column( $image_rows, 'meta_value' ) );
				$all_imgs_ids = array_map( 'intval', explode( ",", $gallery_ids ) );
			}

			// get all images attached to posts and its variations
			$attached_images_args = array(
				'post_parent__in' => $all_ids,
				'post_type'       => 'attachment',
				'numberposts'     => - 1,
				'post_status'     => 'any',
				'post_mime_type'  => 'image',
				'fields'          => 'ids',
			);

			$attached_images = get_posts( $attached_images_args );

			if ( $attached_images ) {
				$all_imgs_ids = array_merge( $attached_images, $all_imgs_ids );
			}

			$all_imgs_ids = array_unique( $all_imgs_ids );

			// all related images with image names
			if ( ! empty( $all_imgs_ids ) ) {
				$all_imgs = $wpdb->get_results( "SELECT p.ID, pm.meta_value as post_title FROM {$wpdb->prefix}posts as p LEFT JOIN `{$wpdb->prefix}postmeta` as pm ON p.ID = pm.post_id WHERE
				pm.meta_key = '_wp_attached_file' and p.ID in (" . implode( ',', $all_imgs_ids ) . ");", ARRAY_A );
				if ( $all_imgs ) {
					// sort by name so we get first uploaded image (without -no)
					usort( $all_imgs, function ( $a, $b ) {
						return ( strlen( $a['post_title'] ) > strlen( $b['post_title'] ) ) ? 1 : - 1;
					} );
				}
			} else {
				$all_imgs = [];
			}

			// checking each image, if image is already attached we need to add it to $all_imgs array to prevent duplicates.
			foreach ( $new_images as $img ) {

				// there is ID, check if that file exists and add it to final array
				if ( isset( $img['id'] ) && wp_attachment_is_image( $img['id'] ) ) {
					$name               = get_post( $img['id'] )->post_title;
					$final_images_ids[] = array(
						'ID' => $img['id'],
//							'post_title' => $name,
					);
					$all_imgs[]         = array(
						'ID'         => $img['id'],
						'post_title' => $name,
					);
					continue;
				}

				// there is source url, check if filename is similar and size is the same.
				if ( ! empty( $img['src'] ) ) {
					$found_img_keys = [];
					// Sanitize filename
					$image_name              = $this->maybe_strip_hash( wp_basename( parse_url( esc_url_raw( $img['src'] ), PHP_URL_PATH ) ) );
					$image_paths             = pathinfo( $image_name );
					$image_paths['filename'] = sanitize_file_name( $image_paths['filename'] );

					// find all keys - name starting with img['src'] filename.
					// TODO - Change to 'starting with' incoming name
					foreach ( $all_imgs as $im ) {
						if ( strpos( $im['post_title'], $image_paths['filename'] ) !== false ) {
							// set first found image if 'ignore_content-length' is set
							if ( isset( $this->settings['ignore_content-length'] ) ) {
								$final_images_ids[] = $im;
								continue 2;
							}
							$found_img_keys[] = $im;
						}
					}

					// found similar name
					if ( $found_img_keys ) {

						// get remote file size
						$remote_file_size = 0;
						$headers          = wp_get_http_headers( $img['src'] );
						if ( $headers ) {
							$headers = $headers->getAll();
							if ( isset( $headers['content-length'] ) ) {
								$remote_file_size = $headers['content-length'];
							}
						}

						// go through all found local images with similar name and check filesize
						foreach ( $found_img_keys as $found_img ) {
							$local_file_size = 0;

							$file = wp_get_original_image_path( $found_img['ID'] );

							if ( file_exists( $file ) ) {
								$local_file_size = filesize( $file );
							}

							if ( $this->sync_compare_content_length( $remote_file_size, $local_file_size ) ) {
								// if name is same and size is same then set existing image id in all image / gallery ids
								$final_images_ids[] = $found_img;
								continue 2;
							}
						}
					}

					// image is not found in local images, upload remote image
					$upload = wc_rest_upload_image_from_url( esc_url_raw( $img['src'] ) );

					if ( is_wp_error( $upload ) ) {
						if ( ! apply_filters( 'woocommerce_rest_suppress_image_upload_error', false, $upload, $object->get_id(), array( $img['src'] ) ) ) {
							throw new WC_REST_Exception( 'woocommerce_product_image_upload_error', $upload->get_error_message(), 400 );
						} else {
							continue;
						}
					}

					$attachment_id = wc_rest_set_uploaded_image_as_attachment( $upload, $object->get_id() );

					// Set the image alt if present.
					if ( ! empty( $img['alt'] ) ) {
						update_post_meta( $attachment_id, '_wp_attachment_image_alt', wc_clean( $img['alt'] ) );
					}

					// Set the image name if present.
					if ( ! empty( $img['name'] ) ) {
						wp_update_post(
							array(
								'ID'         => $attachment_id,
								'post_title' => $img['name'],
							)
						);
					}


					// adding to our all local images array
					$all_imgs[]         = array(
						'ID'         => $attachment_id,
						'post_title' => $image_name,
					);
					$final_images_ids[] = array(
						'ID'         => $attachment_id,
						'post_title' => $image_name,
					);
				}
			}
		}

		if ( $final_images_ids ) {
			$gallery = array();
			foreach ( $final_images_ids as $index => $fid ) {
				if ( 0 === $index ) {
					$object->set_image_id( intval( $fid['ID'] ) );
				} else {
					$gallery[] = intval( $fid['ID'] );
				}
			}

			delete_post_meta( $object->get_ID(), '_sysp_bulk_images' );

			$object->set_gallery_image_ids( $gallery );
			$object->save();

			// cleanup
			wc_delete_product_transients( $object->get_id() );
		}

		return $object;
	}


	/**
	 * @param WC_Data $object Inserted object.
	 * @param WP_REST_Request $request Request object.
	 * @param boolean $creating True when creating object, false when updating.
	 *
	 * @return mixed
	 * @throws WC_REST_Exception
	 */
	public function sysp_update_variation_image( $object, $request, $created ) {
		global $wpdb;

		$new_main_image = $object->get_meta( '_sysp_var_image' );

		if ( $new_main_image ) {

			$all_img_keys = [];
			$all_imgs_ids = [];

			$parent_product = wc_get_product( $object->get_parent_id() );

			// get all variations ids and push parent id to it
			$all_ids   = $parent_product->get_children();
			$all_ids[] = $object->get_parent_id();

			// get ids of gallery images + featured image. Direct query
			$image_rows = $wpdb->get_results( "SELECT `meta_value` FROM `{$wpdb->prefix}postmeta` WHERE (`meta_key` = '_thumbnail_id' OR `meta_key` = '_product_image_gallery') AND `post_id` in (" . implode( ',', $all_ids ) . ");", ARRAY_A );

			if ( $image_rows ) {
				$gallery_ids  = implode( ',', array_column( $image_rows, 'meta_value' ) );
				$all_imgs_ids = array_map( 'intval', explode( ",", $gallery_ids ) );
			}

			// images attached to posts, maybe attached but not in gallery
			$attached_images_args = array(
				'post_parent__in' => $all_ids,
				'post_type'       => 'attachment',
				'numberposts'     => - 1, // show all
				'post_status'     => 'any',
				'post_mime_type'  => 'image',
				'fields'          => 'ids',
			);

			$attached_images = get_posts( $attached_images_args );

			if ( $attached_images ) {
				$all_imgs_ids = array_merge( $attached_images, $all_imgs_ids );
			}

			$all_imgs_ids = array_unique( $all_imgs_ids );

			// there is ID, check if that file exists
			if ( isset( $new_main_image['id'] ) && wp_attachment_is_image( $new_main_image['id'] ) ) {
				$object->set_image_id( (int) $new_main_image['id'] );
				delete_post_meta( $object->get_ID(), '_sysp_var_image' );
				$object->save();
				wc_delete_product_transients( $object->get_ID() );

				return $object;
			}

			// all related images with image names
			if ( ! empty( $all_imgs_ids ) ) {
				$all_imgs = $wpdb->get_results( "SELECT p.ID, pm.meta_value as post_title FROM {$wpdb->prefix}posts as p LEFT JOIN `{$wpdb->prefix}postmeta` as pm ON p.ID = pm.post_id WHERE
				pm.meta_key = '_wp_attached_file' and p.ID in (" . implode( ',', $all_imgs_ids ) . ");", ARRAY_A );
				if ( $all_imgs ) {
					// sort by name so we get first uploaded image (without -no)
					usort( $all_imgs, function ( $a, $b ) {
						return ( strlen( $a['post_title'] ) > strlen( $b['post_title'] ) ) ? 1 : - 1;
					} );
				}
			} else {
				$all_imgs = [];
			}

			//sanitizing image name
			$image_name              = $this->maybe_strip_hash( wp_basename( parse_url( esc_url_raw( $new_main_image['src'] ), PHP_URL_PATH ) ) );
			$image_paths             = pathinfo( $image_name );
			$image_paths['filename'] = sanitize_file_name( $image_paths['filename'] );

			// remove -scaled
			$image_paths['filename'] = preg_replace( '/-scaled$/', '', $image_paths['filename'] );

			// remove number sufix (case: image uploaded more than once, and reasigned to other variation)
			$image_paths['filename'] = preg_replace( '/(-[0-9]{1,3})$/', '', $image_paths['filename'] );

			// finding all keys - maybe we have more than one image with similar name
			// TODO - Change to 'starting with' incoming name
			foreach ( $all_imgs as $im ) {
				if ( strpos( $im['post_title'], $image_paths['filename'] ) !== false ) {
					// set first found image if 'ignore_content-length' is set
					if ( isset( $this->settings['ignore_content-length'] ) ) {
						$object->set_image_id( intval( $im['ID'] ) );
						delete_post_meta( $object->get_ID(), '_sysp_var_image' );
						$object->save();
						wc_delete_product_transients( $object->get_ID() );

						return $object;
					}
					$all_img_keys[] = $im;
				}
			}

			// found same name
			if ( $all_img_keys ) {

				$remote_file_size = 0;

				$headers = wp_get_http_headers( $new_main_image['src'] );
				if ( $headers ) {
					$headers = $headers->getAll();

					if ( isset( $headers['content-length'] ) ) {
						$remote_file_size = $headers['content-length'];
					}
				}


				// go through all found local images with similar name and check filesize
				foreach ( $all_img_keys as $img_key ) {

					$local_file_size = 0;
					$file = wp_get_original_image_path( $img_key['ID'] );

					if ( file_exists( $file ) ) {
						$local_file_size = filesize( $file );
					}

					if ( $this->sync_compare_content_length( $remote_file_size, $local_file_size ) ) {
						// if name is same and size is same then set existing image id
						$object->set_image_id( intval( $img_key['ID'] ) );
						delete_post_meta( $object->get_ID(), '_sysp_var_image' );
						$object->save();
						wc_delete_product_transients( $object->get_ID() );

						return $object;
					}
				}

			}

			$upload = wc_rest_upload_image_from_url( esc_url_raw( $new_main_image['src'] ) );

			if ( is_wp_error( $upload ) ) {
				if ( ! apply_filters( 'woocommerce_rest_suppress_image_upload_error', false, $upload, $object->get_id(), array( $new_main_image['src'] ) ) ) {
					throw new WC_REST_Exception( 'woocommerce_variation_image_upload_error', $upload->get_error_message(), 400 );
				}
			} else {
				$attachment_id = wc_rest_set_uploaded_image_as_attachment( $upload, $object->get_id() );

				// Set the image alt if present.
				if ( ! empty( $new_main_image['alt'] ) ) {
					update_post_meta( $attachment_id, '_wp_attachment_image_alt', wc_clean( $new_main_image['alt'] ) );
				}

				// Set the image name if present.
				if ( ! empty( $new_main_image['name'] ) ) {
					wp_update_post(
						array(
							'ID'         => $attachment_id,
							'post_title' => $new_main_image['name'],
						)
					);
				}

				$object->set_image_id( $attachment_id );
				delete_post_meta( $object->get_id(), '_sysp_var_image' );
				$object->save();
				wc_delete_product_transients( $object->get_id() );
			}
		}

		return $object;
	}


	/**
	 * Compare headers['content-length'] with our local filesize. Works with strings and array.
	 *
	 * @param mixed $remote_content_length
	 * @param string $local_content_length
	 *
	 * @return bool
	 */
	private function sync_compare_content_length( $remote_content_length, $local_content_length ) {
		if ( is_array( $remote_content_length ) ) {
			return ( array_search( $local_content_length, $remote_content_length ) !== false );
		}

		if ( is_numeric( $remote_content_length ) ) {
			return ( $remote_content_length == $local_content_length );
		}

		return false;
	}

	/**
	 * Check if there is hash in filename and strip it.
	 *
	 * @param $filename
	 *
	 * @return false|mixed|string
	 */
	private function maybe_strip_hash( $filename ) {
		$hash = mb_substr( $filename, 0, 33, 'utf-8' );

		if ( strlen( $hash ) === 33 && mb_strlen( $filename ) > 37 && preg_match( '/^[a-f0-9]{32}_$/', $hash ) ) {
			return mb_substr( $filename, 33, null, 'utf-8' );
		}

		return $filename;
	}

}